#ifndef LIBVMMA_H
#define LIBVMMA_H

#include <stddef.h>
#include <stdint.h>


/* opaque type for a VMMA context */
struct vmma_context;


/* opaque type for a VMMA client */
struct vmma_client;


/* create a thread-safe allocator context (locking will performed */
/* on enterring functions, and unlocking on leaving functions */
#define VMMA_FL_MULTITHREADED   0x00000001u


/* grow the client allocation with the minimal amount as possible */
/* (this prevents preallocation, and may reduce performance) */
#define VMMA_FL_MINGROWTH       0x00000002u


/* keep cached memory blocks when the working size is exhausted */
/* (this forces to allocate a new block instead of trying to serve */
/* the allocation request by freeing up the caches memory blocks; */
/* however, this mechanism may reduce internal fragmentation) */
#define VMMA_FL_KEEPCACHED      0x00000004u


/* minimal client align in vmma_create_context() */
#define VMMA_CLIENTALIGN_MIN 16


/* minimal block size in vmma_create_context() */
#define VMMA_BLOCKSIZE_MIN 1024


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* This function returns the supported flags (VMMA_FL_*). */
/* User shall check whether the requested feature is enabled or not. */
unsigned int vmma_get_supported_flags(void);


/* This function creates a new VMMA context. A context may have clients. */
/* Actual memory allocation can be performed using clients. */
struct vmma_context* vmma_create_context(
	/* the memory to work with */
	void* const memory,
	/* size of the memory block */
	const size_t memsize,
	/* desired memory address alignment of memory blocks allocated by */
	/* clients (this must be power of two, cannot be less than */
	/* VMMA_CLIENTALIGN_MIN) */
	const size_t clientalign,
	/* allocation unit size of the original memory block -- this shall */
	/* be a few kilobytes, i.e. 64KiB -- the clients can allocate memory */
	/* blocks that is multiple of this number (must be power of two, */
	/* cannot be less than VMMA_BLOCKSIZE_MIN)*/
	const size_t blocksize,
	/* flags altering the behavior (unsupported features will be ignored) */
	const unsigned int flags);


/* Delete the VMMA context. The memory chunk will not be accessed anymore. */
void vmma_delete_context(
	/* VMMA context created with vmma_create_context() */
	struct vmma_context* const ctx);


/* Get the accepted flags passed in vmma_create_context(). */
unsigned int vmma_get_accepted_flags(
	/* VMMA context created with vmma_create_context() */
	struct vmma_context* const ctx);


/* Get total allocatable memory size in the context. */
/* This value computed during the initialization. */
size_t vmma_get_total_size(
	/* VMMA context created with vmma_create_context() */
	struct vmma_context* const ctx);


/* Get free memory size that can be used to create clients. */
/* Only the free memory of the context will be populated. */
size_t vmma_get_free_size(
	/* VMMA context created with vmma_create_context() */
	struct vmma_context* const ctx);


/* Populate the total free size that can be allocated. */
/* Every single free memory chunk will be included. */
/* This takes some time to populate. */
size_t vmma_populate_free_size(
	/* VMMA context created with vmma_create_context() */
	struct vmma_context* const ctx);


/* Create a new client. The memory will be allocated from the context. */
struct vmma_client* vmma_create_client(
	/* VMMA context created with vmma_create_context() */
	struct vmma_context* const ctx,
	/* limit the memory usage of the client (0 means unlimited) */
	const size_t limit);


/* Delete a client. */
/* The allocations of the client are not needed to be freed up. */
/* This is the point of the whole mechanism is: deal with memory */
/* leaks made by clients. All memory used by the client will be */
/* claimed by the context. */
void vmma_delete_client(
	/* VMMA client created with vmma_create_client() */
	struct vmma_client* const client);


/* Allocate a memory block. */
/* The returned memory block will be aligned to the 'clientalign' */
/* specified in vmma_create_context(). */
void* vmma_alloc(
	/* VMMA client created with vmma_create_client() */
	struct vmma_client* const client,
	/* size to allocate */
	size_t size);


/* Free a previously allocated memory block. */
void vmma_free(
	/* VMMA client created with vmma_create_client() */
	struct vmma_client* const client,
	/* previously allocated memory block using the same client */
	void* memory);


/* Query the size of the allocated memory block. Returns zero if the block */
/* was not allocated inside the specified client. Useful for debugging. */
size_t vmma_get_size(
	/* VMMA client created with vmma_create_client() */
	struct vmma_client* const client,
	/* previously allocated memory block using the same client */
	void* memory);


/* Test a memory address whether it is allocated in the specified client. */
/* Returns the same address on success; NULL otherwise. */
void* vmma_test_address(
	/* VMMA client created with vmma_create_client() */
	struct vmma_client* const client,
	/* previously allocated memory block using the same client */
	void* address);


/* Query the client memory usage limit specified in vmma_create_client(). */
size_t vmma_get_client_limit(
	/* VMMA client created with vmma_create_client() */
	struct vmma_client* const client);


/* Query the client working size. */
/* Returns the total allocated memory size from the context. This also */
/* includes the administration space as well. */
size_t vmma_get_client_working_size(
	/* VMMA client created with vmma_create_client() */
	struct vmma_client* const client);


/* Query the client memory usage. */
/* Returns the total size of the allocated memory using vmma_alloc() on */
/* the specified client. Also includes the alignment roundings. */
size_t vmma_get_client_usage(
	/* VMMA client created with vmma_create_client() */
	struct vmma_client* const client);


/* Return the total size of cached memory blocks. */
/* Not efficient, shall not be called often. */
size_t vmma_populate_cached_blocks(
	/* VMMA client created with vmma_create_client() */
	struct vmma_client* const client);


/* Force to purge cached memory blocks. This may reduce internal */ 
/* fragmentation, and makes a larger allocation possible. */
/* Not efficient, shall not be called often. */
void vmma_purge_client_cache(
	/* VMMA client created with vmma_create_client() */
	struct vmma_client* const client);


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* LIBVMMA_H */
