//================================================================================================
// $Id: cgsection.h 6308 2023-01-16 20:16:41Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Code generator section
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#ifndef WTC_CGSECTION_H
#define WTC_CGSECTION_H

#include <string>
#include <vector>
#include <map>
#include "wtruntime/wtruntime.h"
#include "wtruntime/types.h"
#include "wtruntime/byteorder.h"

namespace wtast
{
class FunctionSignature;
class Type;
class RecordType;
class OperatorDeclaration;
class DataType;
class DataTypeCons;
class EnumType;
class FunctionType;
} // wtast

namespace wtc
{
class ProblemHandler;
} // wtc

namespace wtc
{
namespace codegen
{

class File;
class InstrumentEngine;

typedef std::vector<wt_u8> RawData;

class Section
{
protected:
	wtc::ProblemHandler & mProbHan;

private:
	std::string mName;
	wt_u32      mAddr;

public:
	Section(wtc::ProblemHandler & probhan, const std::string & name,
		wt_u32 addr)
		: mProbHan(probhan)
		, mName(name)
		, mAddr(addr)
		{ }
	virtual ~Section() { }
	virtual wt_u8 GetSectionType() const = 0;
	virtual wt_u8 GetSectionFlag() const = 0;
	virtual const std::string & GetSectionLink() const { return EmptyString; }
	virtual bool IsEmpty() const = 0;
	virtual wt_u32 GetSize() const = 0;
	virtual wt_u32 GetAddr() const { return mAddr; }
	virtual void Write(File & file) const = 0;

	const std::string & GetName() const { return mName; }

	static const std::string EmptyString;
};

template<wt_u8 TYPE, wt_u8 FLAG>
class SectionImpl : public Section
{
public:
	SectionImpl(wtc::ProblemHandler & probhan, const std::string & name,
		wt_u32 addr) : Section(probhan, name, addr) { }
	virtual ~SectionImpl() { }
	virtual wt_u8 GetSectionType() const { return TYPE; }
	virtual wt_u8 GetSectionFlag() const { return FLAG; }
};

class MetadataSection : public SectionImpl<WTBST_METADATA, 0>
{
	typedef std::map<std::string, std::string> metadata_map_t;

	metadata_map_t mMetadata;

public:
	MetadataSection(wtc::ProblemHandler & probhan, const std::string & name)
		: SectionImpl(probhan, name, 0) { }

	virtual bool IsEmpty() const { return mMetadata.empty(); }
	virtual wt_u32 GetSize() const;
	virtual void Write(File & file) const;

	bool Add(const std::string & key, const std::string & value);
};

class IntegritySection : public SectionImpl<WTBST_INTEGRITY, 0>
{
	wt_u16 mHashAlg;
	wt_u32 mHashSize;
	wt_bo  mByteOrder;

public:
	IntegritySection(wtc::ProblemHandler & probhan, const std::string & name,
		wt_u16 hash_alg, wt_u32 hash_size, wt_bo bo)
		: SectionImpl(probhan, name, 0)
		, mHashAlg(hash_alg)
		, mHashSize(hash_size)
		, mByteOrder(bo)
	{
		// ok
	}

	virtual bool IsEmpty() const { return false; }
	virtual wt_u32 GetSize() const { return sizeof(wtb_integhdr) + mHashSize; }
	virtual void Write(File & file) const;
};

class BosSection : public SectionImpl<WTBST_BOS, 0>
{
	wt_u32      mCount2;
	wt_u32      mCount4;
	wt_u32      mCount8;
	wt_u32      mCountS;
	std::string mDataName;

	static wt_u32 MakeSize(wt_u32 n);
	static void WriteSize(File & file, wt_u32 n, wt_u8 base_bos);

public:
	BosSection(wtc::ProblemHandler & probhan, const std::string & name,
		wt_u32 addr, const std::string & data_name, wt_u32 c2, wt_u32 c4,
		wt_u32 c8, wt_u32 cS)
		: SectionImpl(probhan, name, addr)
		, mCount2(c2)
		, mCount4(c4)
		, mCount8(c8)
		, mCountS(cS)
		, mDataName(data_name)
	{
		// ok
	}

	virtual const std::string & GetSectionLink() const { return mDataName; }
	virtual bool IsEmpty() const { return 0==mCount2 && 0==mCount4 && 0==mCount8 && 0==mCountS; }
	virtual wt_u32 GetSize() const;
	virtual void Write(File & file) const;
};

class DataSection : public SectionImpl<WTBST_DATA, WTBSF_LOAD>
{
	typedef std::map<std::string, wt_u32> data_mapping_t;

	data_mapping_t mData2;
	data_mapping_t mData4;
	data_mapping_t mData8;
	data_mapping_t mDataS;
	std::string    mBosName;
	bool           mCompiled;

	void Add(data_mapping_t & mapping, const std::string & data);
	wt_u32 Find(const data_mapping_t & mapping, const std::string & data) const;

public:
	DataSection(wtc::ProblemHandler & probhan, const std::string & name,
		wt_u32 addr, const std::string & bos_name)
		: SectionImpl(probhan, name, addr)
		, mBosName(bos_name)
		, mCompiled(false)
		{ }

	virtual const std::string & GetSectionLink() const { return mBosName; }
	virtual bool IsEmpty() const { return mData2.empty() && mData4.empty() && mData8.empty() && mDataS.empty(); }
	virtual wt_u32 GetSize() const;
	virtual void Write(File & file) const;

	void Add(wt_u16 v) { std::string data(reinterpret_cast<char*>(&v), sizeof(v)); Add(mData2, data); }
	void Add(wt_u32 v) { std::string data(reinterpret_cast<char*>(&v), sizeof(v)); Add(mData4, data); }
	void Add(wt_u64 v) { std::string data(reinterpret_cast<char*>(&v), sizeof(v)); Add(mData8, data); }
	void Add(const std::string & s) { Add(mDataS, s); }

	BosSection* Compile();

	wt_u32 Find(wt_u16 v) const { std::string data(reinterpret_cast<char*>(&v), sizeof(v)); return Find(mData2, data); }
	wt_u32 Find(wt_u32 v) const { std::string data(reinterpret_cast<char*>(&v), sizeof(v)); return Find(mData4, data); }
	wt_u32 Find(wt_u64 v) const { std::string data(reinterpret_cast<char*>(&v), sizeof(v)); return Find(mData8, data); }
	wt_u32 Find(const std::string & s) const { return Find(mDataS, s); }
};

class ExternalsSection : public SectionImpl<WTBST_DATA, WTBSF_CLEAR>
{
public:
	typedef std::map<std::string, wt_u32> map_name_offs_t;
	typedef map_name_offs_t::const_iterator const_offs_iterator;

	typedef std::map<std::string, std::pair<wt_u32, wtast::FunctionSignature*> > map_name_offssig_t;
	typedef map_name_offssig_t::const_iterator const_offssig_iterator;

	typedef std::map<std::string, std::pair<wt_u32, wtast::Type*> > map_name_offstype_t;
	typedef map_name_offstype_t::const_iterator const_offstype_iterator;

	typedef std::map<std::string, wtast::OperatorDeclaration*> map_name_op_t;
	typedef map_name_op_t::const_iterator const_op_iterator;

private:
	bool                mCompiled;
	wt_u32              mCompiledSize;
	map_name_offssig_t  mNameToPFN;
	map_name_offs_t     mNameToExc;
	map_name_offstype_t mNameToType;
	map_name_op_t       mNameToOp;

	void _Add(map_name_offs_t & mapping, const std::string & name);
	void _Add(map_name_offssig_t & mapping, wtast::FunctionSignature* sig);
	void _Add(map_name_offstype_t & mapping, wtast::RecordType* rec);
	void _Add(map_name_offstype_t & mapping, wtast::DataType* dt);
	void _Add(map_name_offstype_t & mapping, wtast::EnumType* et);
	void _Add(map_name_offstype_t & mapping, wtast::FunctionType* ft);
	void _Add(map_name_op_t & mapping, wtast::OperatorDeclaration* op);
	wt_u32 _GetNameAddr(const map_name_offs_t & mapping, const std::string & name) const;
	wt_u32 _GetNameAddr(const map_name_offssig_t & mapping, const std::string & name) const;
	wt_u32 _GetNameOffs(const map_name_offstype_t & mapping, const std::string & name) const;

public:
	ExternalsSection(wtc::ProblemHandler & probhan, const std::string & name,
		wt_u32 addr)
		: SectionImpl(probhan, name, addr)
		, mCompiled(false)
		, mCompiledSize(0)
		{ }

	virtual wt_u8 GetSectionFlag() const
	{
		wt_u8 flags = WTBSF_CLEAR;
		if(!mNameToType.empty()) { flags |= WTBSF_TYPEVEC; }
		return flags;
	}

	virtual void Write(File & file) const { } // Nincs mit kiirni a file-ba!
	virtual wt_u32 GetSize() const { return mCompiledSize; }

	virtual bool IsEmpty() const
	{
		return
			mNameToPFN.empty() &&
			mNameToExc.empty() &&
			mNameToType.empty() &&
			mNameToOp.empty();
	}

	std::size_t GetNumExternals() const
	{
		return
			mNameToPFN.size() +
			mNameToExc.size() +
			mNameToType.size() +
			mNameToOp.size();
	}

	bool IsTypeVec() const { return !mNameToType.empty(); }

	void AddFunc(wtast::FunctionSignature* sig) { _Add(mNameToPFN, sig); }
	void AddExc(const std::string & excname) { _Add(mNameToExc, excname); }
	void AddType(wtast::RecordType* rec) { _Add(mNameToType, rec); }
	void AddType(wtast::DataType* dt) { _Add(mNameToType, dt); }
	void AddType(wtast::EnumType* et) { _Add(mNameToType, et); }
	void AddType(wtast::FunctionType* ft) { _Add(mNameToType, ft); }
	void AddOperator(wtast::OperatorDeclaration* op) { _Add(mNameToOp, op); }

	void Compile();

	static const wt_u32 BadAddr;
	static const wt_u32 BadOffs;

	wt_u32 GetFuncAddr(const std::string & fullname) const { return _GetNameAddr(mNameToPFN, fullname); }
	wt_u32 GetExcAddr(const std::string & excname) const { return _GetNameAddr(mNameToExc, excname); }
	wt_u32 GetTypeOffs(const std::string & name) const { return _GetNameOffs(mNameToType, name); }

	bool HasFunc(const std::string & fullname) const { return mNameToPFN.end()!=mNameToPFN.find(fullname); }
	bool HasType(const std::string & name) const { return mNameToType.end()!=mNameToType.find(name); }
	bool HasException(const std::string & excname) const { return mNameToExc.end()!=mNameToExc.find(excname); }

	const_offssig_iterator func_begin() const { return mNameToPFN.begin(); }
	const_offssig_iterator func_end() const { return mNameToPFN.end(); }
	const_offs_iterator exc_begin() const { return mNameToExc.begin(); }
	const_offs_iterator exc_end() const { return mNameToExc.end(); }
	const_offstype_iterator type_begin() const { return mNameToType.begin(); }
	const_offstype_iterator type_end() const { return mNameToType.end(); }
	const_op_iterator op_begin() const { return mNameToOp.begin(); }
	const_op_iterator op_end() const { return mNameToOp.end(); }
};

class CodeSection : public SectionImpl<WTBST_CODE, WTBSF_LOAD>
{
	typedef std::vector<wt_u16> instructions_t;
	typedef std::map<std::string, wt_u32> map_func_pc_t;

	instructions_t mInstructions;
	map_func_pc_t  mFunctions;

public:
	CodeSection(wtc::ProblemHandler & probhan, const std::string & name,
		wt_u32 addr) : SectionImpl(probhan, name, addr) { }

	virtual bool IsEmpty() const { return mInstructions.empty(); }
	virtual wt_u32 GetSize() const { return mInstructions.size()*sizeof(wt_u16); }
	virtual void Write(File & file) const;

	void AddInstruction(wt_u16 ins) { mInstructions.push_back(ins); }
	void AddFunction(const std::string & name, wt_u32 pc) { mFunctions[name] = pc; }
	wt_u32 GetNumInstructions() const { return mInstructions.size(); }

	wt_u32 GetFunctionPC(const std::string & name) const;

	friend class InstrumentEngine;
};

class GenericImportExportSection : public SectionImpl<0, 0>
{
protected:
	struct Data
	{
		enum Type
		{
			T_VOID,
			T_U8,
			T_U32,
			T_STRING
		};

		Data() : type(T_VOID), v_u8(), v_u32() { }
		Data(wt_u8 v) : type(T_U8), v_u8(v), v_u32() { }
		Data(wt_u32 v) : type(T_U32), v_u8(), v_u32(v) { }
		Data(const std::string & v) : type(T_STRING), v_u8(), v_u32(), v_string(v) { }

		Type        type;
		wt_u8       v_u8;
		wt_u32      v_u32;
		std::string v_string;
	};

	typedef std::vector<wtb_sym> symvec_t;
	typedef std::vector<Data>    datavec_t;

	symvec_t  mSymbols;
	wt_u32    mRelSize;
	datavec_t mData;

	static const wt_u32 Padding = sizeof(wt_u32);

	void PlaceString(const std::string & str);
	void PlaceString(const std::string & str, wt_u32 padding);
	void PlaceFuncSig(wtast::FunctionSignature* sig);
	void PlaceRecord(wtast::RecordType* rec);
	void PlaceDataCons(wtast::DataTypeCons* cons);
	void SYM_DataType(wtast::DataType* dt, wt_u32 sym_info);
	void SYM_EnumType(wtast::EnumType* et, wt_u32 sym_info);
	void SYM_FunctionType(wtast::FunctionType* ft, wt_u32 sym_info);
	void SYM_Operator(wtast::OperatorDeclaration* op);

public:
	GenericImportExportSection(wtc::ProblemHandler & probhan,
		const std::string & name)
		: SectionImpl(probhan, name, 0), mRelSize(0) { }

	virtual bool IsEmpty() const { return mSymbols.empty(); }
	virtual wt_u32 GetSize() const { return sizeof(wtb_symhdr) + sizeof(wtb_sym)*mSymbols.size() + mRelSize; }
	virtual void Write(File & file) const;
};

class ImportSection : public GenericImportExportSection
{
public:
	ImportSection(wtc::ProblemHandler & probhan, const std::string & name)
		: GenericImportExportSection(probhan, name) { }

	virtual wt_u8 GetSectionType() const { return WTBST_IMPORT; }

	void AddFunctionImport(wtast::FunctionSignature* sig, wt_u32 offs);
	void AddExceptionImport(const std::string & exc_name, wt_u32 offs);
	void AddTypeImport(wtast::Type* type, wt_u32 info);
	void AddOperatorImport(wtast::OperatorDeclaration* op);
};

class ExportSection : public GenericImportExportSection
{
public:
	ExportSection(wtc::ProblemHandler & probhan, const std::string & name)
		: GenericImportExportSection(probhan, name) { }

	virtual wt_u8 GetSectionType() const { return WTBST_EXPORT; }

	void AddFunctionExport(wtast::FunctionSignature* sig, wt_u32 pc);
	void AddExceptionExport(const std::string & exc_name);
	void AddTypeExport(wtast::Type* type, wt_u32 sym_info);
	void AddOperatorExport(wtast::OperatorDeclaration* op);
};

class GlobalSection : public SectionImpl<WTBST_DATA, WTBSF_CLEAR>
{
	wt_u32 mSize;

public:
	GlobalSection(wtc::ProblemHandler & probhan, const std::string & name,
		wt_u32 addr, wt_u32 size)
		: SectionImpl(probhan, name, addr), mSize(size) { }

	virtual bool IsEmpty() const { return 0==mSize; }
	virtual wt_u32 GetSize() const { return mSize; }
	virtual void Write(File & file) const { }
};

class DLabelSection : public SectionImpl<WTBST_DEBUGLABEL, 0>
{
	typedef std::map<wt_u32, std::string> map_addr_label_t;

	map_addr_label_t mMapAddrToLabel;

public:
	DLabelSection(wtc::ProblemHandler & probhan, const std::string & name)
		: SectionImpl(probhan, name, 0) { }

	virtual bool IsEmpty() const { return mMapAddrToLabel.empty(); }
	virtual wt_u32 GetSize() const;
	virtual void Write(File & file) const;

	void AddLabel(wt_u32 addr, const std::string & label) { mMapAddrToLabel[addr] = label; }
};

} // codegen
} // wtc

#endif // WTC_CGSECTION_H
