//================================================================================================
// $Id: codegen.h 6308 2023-01-16 20:16:41Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Code generator
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#ifndef WTC_CODEGEN_H
#define WTC_CODEGEN_H

#include <map>
#include <set>
#include <vector>
#include <string>
#include "cgselector.h"
#include "cginstrument.h"
#include "wtruntime/wtruntime.h"
#include "wtparser/uniqueid.h"

#define AddAssert(_probhan_, _expr_) \
	(!(_expr_) && AddAssert_impl(_probhan_, __FILE__, __LINE__, #_expr_))

namespace wtast
{
class AstNode;
class Program;
class Expression;
class FunctionDefinition;
class FunctionSignature;
class Block;
class MainBlock;
class ConstructorBlock;
class DestructorBlock;
class Type;
class TypeInfo;
class MakeFunctionPointer;
} // wtast

namespace wtc
{
class FrameStack;
class Options;
class ProblemHandler;
bool AddAssert_impl(ProblemHandler* const probhan, const char* const file,
	unsigned int line, const char* const expr);
} // wtc

namespace wtc
{
namespace codegen
{

struct Exception
{
	Exception(const std::string & msg) : msg(msg) { }
	const std::string msg;
};

class Section;
class DataSection;
class ExternalsSection;
class ImportSection;
class ExportSection;
class CollectVarAccessVisitor;

class CodeGen
{
	struct FuncnameType
	{
		std::string  funcname;
		wtast::Type* type;

		FuncnameType() : type(0) { }
		FuncnameType(const std::string & n, wtast::Type* t) : funcname(n), type(t) { }
	};

	typedef wtast::FunctionCall        ast_FC;
	typedef wtast::MakeFunctionPointer ast_MKFP;

	typedef void (CodeGen::*PfnGenFunc)(wtast::FunctionSignature* sig);

	typedef std::vector<wtast::AstNode*>                              astnodevec_t;
	typedef std::map<std::string, Section*>                           sections_t;
	typedef std::vector<wtast::FunctionSignature*>                    funcsigvec_t;
	typedef std::vector<wtast::Type*>                                 typevec_t;
	typedef std::set<std::string>                                     stringset_t;
	typedef std::set<wtast::FunctionDefinition*>                      funcdefset_t;
	typedef std::map<wtparser::UniqueId, wtast::FunctionDefinition*>  map_uniqueid_funcdef_t;
	typedef std::map<wtparser::UniqueId, wtast::FunctionSignature*>   map_uniqueid_funcsig_t;
	typedef std::map<std::string, FuncnameType>                       map_typename_functype_t;
	typedef Selector<wtast::Expression, true, ast_FC, ast_MKFP>       FuncCallSelector;
	typedef Selector<wtast::GlobalBlock, false>                       GlobalBlockSelector;
	typedef std::set<wtparser::UniqueId>                              uniqueidset_t;
	typedef std::vector<wtast::VariableDeclFragment*>                 vardeclfrags_t;
	typedef std::map<wtparser::UniqueId, wtast::VariableDeclaration*> map_uniqueid_vardecl_t;
	typedef std::set<wtast::VariableDeclaration*>                     vardecls_t;
	typedef std::map<std::string, PfnGenFunc>                         map_str_genfunc_t;

	wtc::Options*             mOptions;
	wtc::ProblemHandler*      mProblemHandler;
	InstrumentProblemHandler* mInstrumentProblemHandler;
	wtast::Program*           mProgram;
	wtparser::UniqueId        mNextUniqueId;
	sections_t                mSections;
	unsigned int              mNextId;
	wt_u32                    mMainPC;
	wt_u32                    mCtorPC;
	wt_u32                    mDtorPC;
	wt_u32                    mDataAddress;   // data kezdocime
	wt_u32                    mExtAddress;    // externals kezdocime
	wt_u32                    mGlobalAddress; // globalis valtozok helye
	wt_u32                    mCodeAddress;   // programkod kezdocime
	wt_u32                    mProgramSize;   // program merete
	wt_u32                    mGlobalSize;    // globalis valtozok merete
	wtb_thdr                  mMainThread;
	funcdefset_t              mFunctionDefs;  // azok a fuggveny definiciok, amikbol kodot kell generalni
	map_uniqueid_funcsig_t    mUidToFuncSig;  // function UniqueId mappelese FunctionSignature-re
	map_uniqueid_funcdef_t    mUidToFuncDef;  // function UniqueId mappelese FunctionDefinition-re
	map_typename_functype_t   mCtorNames;
	map_typename_functype_t   mDtorNames;
	map_typename_functype_t   mCopyNames;
	map_uniqueid_funcsig_t    mGeneratedFunctionNames;
	stringset_t               mGeneratedFuncImplNames;
	stringset_t               mUnconstructibleTypes;
	InstrumentEngine          mEngine;
	wtast::MainBlock*         mMainBlock;
	wtast::ConstructorBlock*  mCtorBlock;
	wtast::DestructorBlock*   mDtorBlock;
	ExternalsSection*         mExternals;
	funcsigvec_t              mExportedFunctions;
	uniqueidset_t             mExportedFunctionIds; // with UniqueId
	uniqueidset_t             mImportedFunctionIds; // with UniqueId
	stringset_t               mExternalTypeSet;
	stringset_t               mImportedTypeSet;
	stringset_t               mExportedTypeSet;
	stringset_t               mDefinedFunctions;
	stringset_t               mKnownOperators;
	astnodevec_t              mAllocatedAstNodes;
	uniqueidset_t             mCalledFunctions; // with UniqueId
	vardeclfrags_t            mGlobalVariables;
	wtast::Type*              mUnifiedObjectType;
	wtast::Type*              mIndexTypeU32;
	map_uniqueid_vardecl_t    mGlobFragToGlobDecl;
	vardecls_t                mReferredGlobDecls;
	map_str_genfunc_t         mGenFuncs;

public:
	static const unsigned int StartAddress   = 0x8;
	static const unsigned int NumberWidth    = 0x8;
	static const          int SectionPadding = 0x4;
	static const std::string  MainFuncName;
	static const std::string  CtorFuncName;
	static const std::string  DtorFuncName;
	static const std::string  GlobalCtorFuncName;
	static const std::string  NewObjFuncName;
	static const std::string  CopyObjFuncName;
	static const std::string  DelObjFuncName;
	static const std::string  GeneratedFuncTag;
	static const std::string  AssertionFailExc;
	static const std::string  PrecondViolExc;
	static const std::string  PostcondViolExc;
	static const std::string  ConstraintErrExc;
	static const std::string  PlaceholderEmptyString;

public:
	CodeGen(wtc::Options* options, wtc::ProblemHandler* probhan, wtast::Program* program);

	~CodeGen();
	const wtc::Options* GetOptions() const { return mOptions; }
	void Save(const std::string & filename);
	std::string GenName(const std::string & name);
	const std::string & GetConstructorName(wtast::TypeInfo* typeinfo);
	const std::string & GetDestructorName(wtast::TypeInfo* typeinfo);
	const std::string & GetCopyName(wtast::TypeInfo* typeinfo);
	void GenConstructorCall(wtast::Type* type);
	void GenCopyCall(wtast::Type* type);
	void GenDestructorCall(wtast::Type* type);
	void GenDestructorMkfp(wtast::Type* type);
	void GenGenericCall(const std::string & fullname);
	void GenGenericMkfp(const std::string & fullname);
	void RegisterGeneratedFunction(wtast::FunctionSignature* sig);
	bool IsFunctionDefined(const std::string & fullname, bool add_as_definition);
	wt_u32 GetGlobalAddress() const { return mGlobalAddress; }
	void AddGlobalVariable(wtast::VariableDeclFragment* frag, wt_u32 curr_dvt_size);
	void FillGlobalVariables(FrameStack & framestack);
	wtparser::UniqueId GetNextUniqueId();
	wtast::Type* GetUnifiedObjectType();
	wtast::Type* GetIndexTypeU32();
	void ValidateFunctionDefinition(wtast::FunctionDefinition* def);
	void DefaultInitializeRecordField(wtast::RecordType* rec, wt_u32 pvt_offs, wtast::RecordField* field);
	bool IsReferredGlobal(wtast::VariableDeclaration* vardecl);

	void RegisterSection(Section* sec);

	void AddNote(wtast::AstNode* astnode, const std::string & msg);
	void AddWarning(wtast::AstNode* astnode, const std::string & msg);
	void AddError(wtast::AstNode* astnode, const std::string & msg);

	void AddNote(const std::string & file, unsigned int line, const std::string & msg);
	void AddWarning(const std::string & file, unsigned int line, const std::string & msg);
	void AddError(const std::string & file, unsigned int line, const std::string & msg);

	static wt_u32 AlignSizeUp(wt_u32 size);

	typedef void (CodeGen::*GenFunc_FT)(FuncnameType & ft);
	typedef void (CodeGen::*GenFunc_S)(wtast::FunctionSignature* sig);

	friend class CollectVarAccessVisitor;
	friend class GenVisitor;

private:
	void ProcessMetadata();
	void PopulateFunctions();
	void ProcessDefinitions(funcdefset_t & unprocessed_definitions);
	void ProcessFuncCalls(funcdefset_t & unprocessed_definitions, FuncCallSelector & calls);
	ExternalsSection* CollectDeclarations();
	DataSection* CollectData();
	void CollectData(DataSection* ds, wtast::AstNode* root);
	void CollectGlobals(wtast::AstNode* root);
	void GenCtor(FuncnameType & ft);
	void GenDtor(FuncnameType & ft);
	void GenCopy(FuncnameType & ft);
	void GenSeqCtor(const std::string & global_label, wtast::SeqType* seqtype, wtast::Type* elemtype);
	void GenSeqDtor(const std::string & global_label, wtast::SeqType* seqtype, wtast::Type* elemtype);
	void GenSeqCopy(const std::string & global_label, wtast::SeqType* seqtype, wtast::Type* elemtype);
	void GenGeneratedFunc(wtast::FunctionSignature* sig);
	void GenSeq_empty(wtast::FunctionSignature* sig);
	void GenSeq_seqlen(wtast::FunctionSignature* sig);
	void GenSeq_seqfill(wtast::FunctionSignature* sig);
	void GenSeq_subseq(wtast::FunctionSignature* sig);
	void GenSeq_lov(wtast::FunctionSignature* sig);
	void GenSeq_hiv(wtast::FunctionSignature* sig);
	void GenSeq_lorem(wtast::FunctionSignature* sig);
	void GenSeq_hirem(wtast::FunctionSignature* sig);
	void GenSeq_loext(wtast::FunctionSignature* sig);
	void GenSeq_hiext(wtast::FunctionSignature* sig);
	void GenSeq_lopop(wtast::FunctionSignature* sig);
	void GenSeq_hipop(wtast::FunctionSignature* sig);
	void GenSeq__copy_loop(wtast::Type* seqtype, wtast::Type* elemtype, wt_u32 offs_src, wt_u32 offs_dst, wt_u32 offs_i_from, wt_u32 offs_i_to, wt_u32 offs_n, wt_u32 add_i_from, wt_u32 add_i_to);
	void GenHelp__incr(wt_u32 offs);
	void GenHelp__decr(wt_u32 offs);
	void GenHelp__decr_if_nonzero(wt_u32 offs);
	void GenIO_FileDtor();
	void GenIO_FileCopy();
	void GenIO_DirDtor();
	void GenIO_DirCopy();
	void GenIO__GenericDtor(wt_u32 exec_code);
	void GenIO__GenericCopy();
	void GenIO__GenericFread(wtast::FunctionSignature* sig);
	void GenIO__GenericFwriteString(wtast::FunctionSignature* sig, wt_char ch);
	void GenIO_fstat(wtast::FunctionSignature* sig);
	void GenIO_fopen(wtast::FunctionSignature* sig);
	void GenIO_feof(wtast::FunctionSignature* sig);
	void GenIO_fread(wtast::FunctionSignature* sig);
	void GenIO_freadchar(wtast::FunctionSignature* sig);
	void GenIO_freadbool(wtast::FunctionSignature* sig);
	void GenIO_freadstr(wtast::FunctionSignature* sig);
	void GenIO_freadln(wtast::FunctionSignature* sig);
	void GenIO_freadi8(wtast::FunctionSignature* sig);
	void GenIO_freadu8(wtast::FunctionSignature* sig);
	void GenIO_freadi16(wtast::FunctionSignature* sig);
	void GenIO_freadu16(wtast::FunctionSignature* sig);
	void GenIO_freadi32(wtast::FunctionSignature* sig);
	void GenIO_freadu32(wtast::FunctionSignature* sig);
	void GenIO_freadf32(wtast::FunctionSignature* sig);
	void GenIO_freadi64(wtast::FunctionSignature* sig);
	void GenIO_freadu64(wtast::FunctionSignature* sig);
	void GenIO_freadf64(wtast::FunctionSignature* sig);
	void GenIO_fwrite(wtast::FunctionSignature* sig);
	void GenIO_fwriteln(wtast::FunctionSignature* sig);
	void GenIO_opendir(wtast::FunctionSignature* sig);
	void GenIO_readdir(wtast::FunctionSignature* sig);
	void GenIO_dirgetname(wtast::FunctionSignature* sig);
	void GenIO_dirgettype(wtast::FunctionSignature* sig);
	void GenStdlib_strtok(wtast::FunctionSignature* sig);
	void GenEnum_succ(wtast::FunctionSignature* sig);
	void GenEnum_pred(wtast::FunctionSignature* sig);
	void GenEnum_ordinal(wtast::FunctionSignature* sig);
	void GenEnum_tos(wtast::FunctionSignature* sig);
	void GenEnum_toEnum_u32(wtast::FunctionSignature* sig);
	void GenEnum_toEnum_string(wtast::FunctionSignature* sig);
	void GenEnum_eq(wtast::FunctionSignature* sig);
	void GenEnum_ne(wtast::FunctionSignature* sig);
	void GenEnum_lt(wtast::FunctionSignature* sig);
	void GenEnum_le(wtast::FunctionSignature* sig);
	void GenEnum_gt(wtast::FunctionSignature* sig);
	void GenEnum_ge(wtast::FunctionSignature* sig);
	void GenEnum__toEnum_string__impl(wtast::EnumType* enum_type, std::string prefix, const std::set<std::string> & remain, const std::string & lab_error);
	void TemplateInstantiator(GenFunc_FT genfunc, map_typename_functype_t & source);
	void TemplateInstantiator(GenFunc_S genfunc, map_uniqueid_funcsig_t & source);
	void GenGenericDtorAndLeave(wtast::Type* type);
	wtast::FunctionSignature* MakeCtorFuncSig(wtast::Type* type);
	wtast::FunctionSignature* MakeCopyFuncSig(wtast::Type* type);
	wtast::FunctionSignature* MakeDtorFuncSig(wtast::Type* type);
	wtast::FunctionSignature* MakeFunctionSignature(const std::string & name, bool pure, wtast::Type* arg, wtast::Type* ret);
	bool IsSafeRegisterAstNode(wtast::AstNode* node, const std::string & alloc_mode_string);
	bool IsSafeAstNode(wtast::AstNode* node, const std::string & alloc_mode_string);
	std::string CreateImplName(const std::string & funcname, wt_u32 loadsize);
	wtast::FunctionSignature* ExtractFunctionSignature(wtast::Expression* expr);
};

} // codegen
} // wtc

#endif // WTC_CODEGEN_H
