#include "config_private.h"
#include "iocall.h"
#include "ctx.h"
#include "stdlib.h"
#include "objects.h"
#include "profile.h"

static const wt_u32 iocall_retv___none[] = { 0 };
static const wt_u32 iocall_retv_u32[] = { WTT_U32 };
static const wt_u32 iocall_retv_i64[] = { WTT_I64 };
static const wt_u32 iocall_retv_bool[] = { WTT_BOOL };
static const wt_u32 iocall_retv_bool_char_u64_u64_u64[] = { WTT_BOOL,
	WTT_CHAR, WTT_U64, WTT_U64, WTT_U64 };
static const wt_u32 iocall_retv_file[] = { WTT_FILE };
static const wt_u32 iocall_retv_dir[] = { WTT_DIR };
static const wt_u32 iocall_retv_char[] = { WTT_CHAR };

static enum wtr_runresult iocall_make_string_from_return(
	struct wtr_context* const ctx)
{
	struct wtr_program* const prog = ctx->frame.prog;

	wt_u32              len;
	wt_vmptr            vma;
	struct wtr_str*     str;
	struct wtr_objentry ent;

	if(WTR_CTX_DS_SIZE(ctx)<sizeof(wt_u32) ||
		WTR_CTX_DS_SPACE(ctx)<sizeof(wt_vmptr))
	{
		ctx->last_exception = StackOverflowException;
		return WTRR_RAISE;
	}

	len = *(wt_u32*)ctx->dsp;
	ctx->dsp = ctx->dsp - sizeof(wt_u32) + sizeof(wt_vmptr);

	if(NULL==(str=wtr_str_create(prog, (char*)ctx->iocall_npriv, len,
		WT_FALSE)))
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	vma = WTR_OBJ_TO_VMADDR(prog, str);
	*(wt_vmptr*)ctx->dsp = vma;

	ent.obj     = vma;
	ent.pfn.ptr = prog;
	ent.pfn.ext = ctx->profile->pc_strd_leave;

	if(!wtr_dynstack_push(&ctx->tos, &ent))
	{
		wtr_str_delete(prog, str);
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

void wtr_iocall_writestdout(struct wtr_context* const ctx, const char* str,
	wt_u32 len)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_WRITESTDOUT;

	ctx->iocall_argc = 2;
	ctx->iocall_retc = 0;
	ctx->iocall_retv = iocall_retv___none;

	arg->writable_type = WTT_STRING;
	arg->v_string      = (char*)str;
	++arg;
	arg->writable_type = WTT_U32;
	arg->v_u32         = len;

	ctx->iocall_next = NULL;
}

void wtr_iocall_readlnstdin(struct wtr_context* const ctx, char sep)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_READLNSTDIN;

	ctx->iocall_argc = 3;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_u32;

	arg->writable_type = WTT_STRING;
	arg->v_string      = ctx->sbuf;
	++arg;
	arg->writable_type = WTT_U32;
	arg->v_u32         = ctx->sbuf_size - 1;
	++arg;
	arg->writable_type = WTT_CHAR;
	arg->v_char        = (wt_u8)sep;

	ctx->iocall_npriv = ctx->sbuf;
	ctx->iocall_next  = iocall_make_string_from_return;
}

void wtr_iocall_access(struct wtr_context* const ctx, const char* pathname,
	wt_u32 mode)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_ACCESS;

	ctx->iocall_argc = 2;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_bool;

	arg->writable_type = WTT_STRING;
	arg->v_string      = (char*)pathname;
	++arg;
	arg->writable_type = WTT_U32;
	arg->v_u32         = mode;

	ctx->iocall_next = NULL;
}

void wtr_iocall_stat(struct wtr_context* const ctx, const char* pathname)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_STAT;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 5;
	ctx->iocall_retv = iocall_retv_bool_char_u64_u64_u64;

	arg->writable_type = WTT_STRING;
	arg->v_string      = (char*)pathname;

	ctx->iocall_next = NULL;
}

void wtr_iocall_fstat(struct wtr_context* const ctx, wt_file file)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FSTAT;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 5;
	ctx->iocall_retv = iocall_retv_bool_char_u64_u64_u64;

	arg->writable_type = WTT_FILE;
	arg->v_file        = file;

	ctx->iocall_next = NULL;
}

void wtr_iocall_fopen(struct wtr_context* const ctx, const char* pathname,
	char mode)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FOPEN;

	ctx->iocall_argc = 2;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_file;

	arg->writable_type = WTT_STRING;
	arg->v_string      = (char*)pathname;
	++arg;
	arg->writable_type = WTT_CHAR;
	arg->v_u32         = mode;

	ctx->iocall_next = NULL;
}

void wtr_iocall_fclose(struct wtr_context* const ctx, wt_file file)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FCLOSE;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 0;
	ctx->iocall_retv = iocall_retv___none;

	arg->writable_type = WTT_FILE;
	arg->v_file        = file;

	ctx->iocall_next = NULL;
}

void wtr_iocall_fseek(struct wtr_context* const ctx, wt_file file,
	wt_i64 offset, wt_u32 whence)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FSEEK;

	ctx->iocall_argc = 3;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_i64;

	arg->writable_type = WTT_FILE;
	arg->v_file        = file;
	++arg;
	arg->writable_type = WTT_I64;
	arg->v_i64         = offset;
	++arg;
	arg->writable_type = WTT_U32;
	arg->v_u32         = whence;

	ctx->iocall_next = NULL;
}

void wtr_iocall_feof(struct wtr_context* const ctx, wt_file file)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FEOF;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_bool;

	arg->writable_type = WTT_FILE;
	arg->v_file        = file;

	ctx->iocall_next = NULL;
}

static enum wtr_runresult iocall_fwrite_checkreturn(
	struct wtr_context* const ctx)
{
	wt_u32 act_len;

	if(WTR_CTX_DS_SIZE(ctx)<sizeof(wt_u32))
	{
		ctx->last_exception = StackOverflowException;
		return WTRR_RAISE;
	}

	act_len = *(wt_u32*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_u32);

	if(act_len!=(wt_uintptr)ctx->iocall_npriv)
	{
		ctx->last_exception = IOException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

static enum wtr_runresult iocall_fwriteln_followup(
	struct wtr_context* const ctx)
{
	const enum wtr_runresult rr = iocall_fwrite_checkreturn(ctx);

	if(WTRR_RUN!=rr || 0==ctx->iocall_wlen)
	{
		return rr;
	}

	wtr_iocall_fwrite(ctx, ctx->iocall_argv[0].v_file, ctx->iocall_wbuf,
		ctx->iocall_wlen, 1);

	/* clear the 'wbuf' */
	ctx->iocall_wlen = 0;

	return WTRR_IOCALL;
}

void wtr_iocall_fwrite(struct wtr_context* const ctx, wt_file file,
	const char* buf, wt_u32 len, int pop_and_check_return_value)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FWRITE;

	ctx->iocall_argc = 3;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_u32;

	arg->writable_type = WTT_FILE;
	arg->v_file        = file;
	++arg;
	arg->writable_type = WTT_STRING;
	arg->v_string      = (char*)buf;
	++arg;
	arg->writable_type = WTT_U32;
	arg->v_u32         = len;

	if(pop_and_check_return_value)
	{
		ctx->iocall_npriv = (void*)(wt_uintptr)len;
		ctx->iocall_next  = iocall_fwrite_checkreturn;
	}
	else
	{
		ctx->iocall_next = NULL;
	}
}

void wtr_iocall_fwriteln(struct wtr_context* const ctx, wt_file file,
	const char* buf, wt_u32 len, char sep)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FWRITE;

	ctx->iocall_argc = 3;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_u32;

	arg->writable_type = WTT_FILE;
	arg->v_file        = file;
	++arg;
	arg->writable_type = WTT_STRING;
	arg->v_string      = (char*)buf;
	++arg;
	arg->writable_type = WTT_U32;
	arg->v_u32         = len;

	ctx->iocall_wlen    = 1;
	ctx->iocall_wbuf[0] = sep;

	ctx->iocall_npriv = (void*)(wt_uintptr)len;
	ctx->iocall_next  = iocall_fwriteln_followup;
}

static enum wtr_runresult iocall_fread_checkreturn(
	struct wtr_context* const ctx)
{
	wt_u32 act_len;

	if(WTR_CTX_DS_SIZE(ctx)<sizeof(wt_u32))
	{
		ctx->last_exception = StackOverflowException;
		return WTRR_RAISE;
	}

	act_len = *(wt_u32*)ctx->dsp;
	WTR_CTX_DS_POP(ctx, wt_u32);

	if(act_len!=(wt_uintptr)ctx->iocall_npriv)
	{
		ctx->last_exception = EndOfFileException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

void wtr_iocall_fread(struct wtr_context* const ctx, wt_file file,
	char* buf, wt_u32 maxlen, int pop_and_check_return_value)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FREAD;

	ctx->iocall_argc = 3;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_u32;

	arg->writable_type = WTT_FILE;
	arg->v_file        = file;
	++arg;
	arg->writable_type = WTT_STRING;
	arg->v_string      = buf;
	++arg;
	arg->writable_type = WTT_U32;
	arg->v_u32         = maxlen;

	if(pop_and_check_return_value)
	{
		ctx->iocall_npriv = (void*)(wt_uintptr)maxlen;
		ctx->iocall_next  = iocall_fread_checkreturn;
	}
	else
	{
		ctx->iocall_next = NULL;
	}
}

void wtr_iocall_freadln(struct wtr_context* const ctx, wt_file file,
	char* buf, wt_u32 maxlen, char sep)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_FREADLN;

	ctx->iocall_argc = 4;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_u32;

	arg->writable_type = WTT_FILE;
	arg->v_file        = file;
	++arg;
	arg->writable_type = WTT_STRING;
	arg->v_string      = buf;
	++arg;
	arg->writable_type = WTT_U32;
	arg->v_u32         = maxlen;
	++arg;
	arg->writable_type = WTT_CHAR;
	arg->v_char        = sep;

	ctx->iocall_npriv = buf;
	ctx->iocall_next  = iocall_make_string_from_return;
}

void wtr_iocall_opendir(struct wtr_context* const ctx, const char* pathname)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_OPENDIR;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_dir;

	arg->writable_type = WTT_STRING;
	arg->v_string      = (char*)pathname;

	ctx->iocall_next = NULL;
}

void wtr_iocall_closedir(struct wtr_context* const ctx, wt_dir dir)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_CLOSEDIR;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 0;
	ctx->iocall_retv = iocall_retv___none;

	arg->writable_type = WTT_DIR;
	arg->v_dir         = dir;

	ctx->iocall_next = NULL;
}

void wtr_iocall_readdir(struct wtr_context* const ctx, wt_dir dir)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_READDIR;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_bool;

	arg->writable_type = WTT_DIR;
	arg->v_dir         = dir;

	ctx->iocall_next = NULL;
}

void wtr_iocall_dirgettype(struct wtr_context* const ctx, wt_dir dir)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_DIRGETTYPE;

	ctx->iocall_argc = 1;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_char;

	arg->writable_type = WTT_DIR;
	arg->v_dir         = dir;

	ctx->iocall_next = NULL;
}

void wtr_iocall_dirgetname(struct wtr_context* const ctx, wt_dir dir,
	char* buf, wt_u32 maxlen)
{
	wt_value* arg = ctx->iocall_argv;

	ctx->last_iocall = WTIOC_DIRGETNAME;

	ctx->iocall_argc = 3;
	ctx->iocall_retc = 1;
	ctx->iocall_retv = iocall_retv_u32;

	arg->writable_type = WTT_DIR;
	arg->v_dir         = dir;
	++arg;
	arg->writable_type = WTT_STRING;
	arg->v_string      = buf;
	++arg;
	arg->writable_type = WTT_U32;
	arg->v_u32         = maxlen;

	ctx->iocall_npriv = buf;
	ctx->iocall_next  = iocall_make_string_from_return;
}
